#
# $Id: kk_style.rb,v 1.4 2004/08/09 09:23:16 kazuki Exp $
#
# tDiary style class for tDiary 2.x format.
# if you want to use this style, add @style into tdiary.conf below:
#
#    @style = 'kk'
#
#
# Copyright (C) 2004, Kazuki Tsujimoto
# You can redistribute it and/or modify it under GPL2.

#  BitChannel: Copyright (C) 2003,2004 Minero Aoki
#  tDiary    : Copyright (C) 2001-2004, TADA Tadashi <sho@spc.gr.jp>

require 'stringio'
require 'uri'

module TDiary
  class KKStyle
    def initialize(body)
      @body = body
    end

    def compile
      @f = LineInput.new(StringIO.new(@body))
      @indent_stack = [0]
      @internal_links = []
      @result = ''
      do_compile
      @result
    end

    #######
    private
    #######

    #
    # Block
    #

    CAPTION = /\A=(?:[^=]|)/
    HEADER = /\A={2,4}/
    UL = /\A\s*\*/
    OL = /\A\s*\(\d+\)|\A\s*\d+\./   # should not allow spaces before '#'
    DL = /\A:/
    CITE = /\A""/
    TABLE = /\A\|\|/
    INDENTED = /\A\s+\S/
    ERB_BLOCK = /\A<%=.*?%>\z/

    PARAGRAPH_END = Regexp.union(CAPTION, HEADER, UL, OL, DL,
      CITE, TABLE, INDENTED, ERB_BLOCK)

    def do_compile
      while @f.next?
        case @f.peek
        when CAPTION, HEADER then caption @f.gets
        when UL              then ul
        when OL              then ol
        when DL              then dl
        when CITE            then cite
        when TABLE           then table
        when INDENTED        then indented_pre   # must be placed after UL/OL
        when ERB_BLOCK       then erb
        else
          if @f.peek.strip.empty?
            @f.gets
            next
          end
          paragraph
        end
      end
    end

    def caption(line)
      head = line.slice(/\A(=+)/, 1)
      level = head.length + 2
      str = line.sub(/\A=+/, '').strip
      puts "<h#{level}>#{escape_html(str)}</h#{level}>"
    end

    def paragraph
      print '<p>'
      nl = ''
      @f.until_match(PARAGRAPH_END) do |line|
        break if line.strip.empty?
        print nl + text(line.sub(/\A\~/, '').strip)
        nl = "\n"
      end
      puts '</p>'
    end

    def ul
      xlist 'ul', UL
    end

    def ol
      xlist 'ol', OL
    end

    LI_CONTINUE = {
      'ul' => /\A\s+[^\s\*]/,
      'ol' => /\A\s+(?!\(\d+\)|\d+\.)\S/
    }

    def xlist(type, mark_re)
      puts "<#{type}>"
      push_indent(indentof(@f.peek)) {
        @f.while_match(mark_re) do |line0|
          line = line0
          if indent_shallower?(line)
            @f.ungets line
            break
          end
          if indent_deeper?(line)
            @f.ungets line
            xlist type, mark_re
            puts '</li>'
            next
          end
          buf = line.sub(mark_re, '').strip
          @f.while_match(LI_CONTINUE[type]) do |line|
            buf << "\n" + line.strip
          end
          if @f.next? and next_line_is_nested_list?(mark_re)
            puts "<li>#{text(buf)}"
          else
            puts "<li>#{text(buf)}</li>"
          end
        end
      }
      puts "</#{type}>"
    end

    def next_line_is_nested_list?(mark_re)
      line = @f.peek
      mark_re =~ line and indent_deeper?(line)
    end

    def dl
      puts '<dl>'
      @f.while_match(DL) do |line|
        if /\A:|\A\s*\z/ =~ @f.peek.to_s
          # original wiki style
          _, dt, dd = line.strip.split(/\s*:\s*/, 3)
          puts "<dt>#{text(dt)}</dt><dd>#{text(dd.to_s)}</dd>"
        else
          # RD style
          dt = line.sub(DL, '').strip
          dd = ''
          @f.while_match(/\A\s+\S/) do |line|
            dd << line.strip << "\n"
          end
          puts "<dt>#{text(dt)}</dt>\n<dd>#{text(dd.strip)}</dd>"
        end
      end
      puts '</dl>'
    end

    def indented_pre
      buf = []
      @f.until_match(/\A\S/) do |line|
        buf.push line
      end
      while buf.last.strip.empty?
        buf.pop
      end
      minindent = buf.map {|line| indentof(line) }.reject {|i| i == 0 }.min
      puts '<pre>'
      buf.each do |line|
        if line.strip.empty?
          puts ''
        else
          puts escape_html(unindent(line.rstrip, minindent))
        end
      end
      puts '</pre>'
    end

    def cite
      puts '<blockquote>'
      print '<p>'
      nl = ''
      @f.while_match(CITE) do |line|
        content = line.sub(CITE, '').strip
        if content.empty?
          print "</p>\n<p>"
          nl = ''
        else
          print nl + escape_html(content)
          nl = "\n"
        end
      end
      puts '</p>'
      puts '</blockquote>'
    end

    def table
      buf = []
      @f.while_match(/\A\|\|/) do |line|
        cols = line.strip.split(/(\|\|\|?)/, -1)
        cols.shift   # discard first ""
        row = []
        until cols.empty?
          isheader = (cols.shift == '|||')
          row.push [cols.shift, isheader]
        end
        buf.push row
      end
      output_table adjust_ncols(buf)
    end

    def adjust_ncols(rows)
      rows.each do |cols|
        while cols.last and cols.last[0].strip.empty?
          cols.pop
          cols.pop
        end
      end
      n_maxcols = rows.map {|cols| cols.size }.max
      rows.each do |cols|
        cols.concat [['',false]] * (n_maxcols - cols.size)
      end
      rows
    end

    def output_table(rows)
      puts '<table>'
      rows.each do |cols|
        puts '<tr>' +
        cols.map {|col, isheader|
          if isheader
          then "<th>#{text(col.strip)}</th>"
          else "<td>#{text(col.strip)}</td>"
          end
        }.join('') +
        '</tr>'
      end
      puts '</table>'
    end

    def erb
      puts @f.gets
    end
    #
    # Indent
    #

    def push_indent(n)
      raise "shollower indent pushed: #{@indent_stack.inspect}" \
        unless n >= current_indent()
      @indent_stack.push(n)
      yield
    ensure
      @indent_stack.pop
    end

    def current_indent
      @indent_stack.last
    end

    def indent_deeper?(line)
      indentof(line) > current_indent()
    end

    def indent_shallower?(line)
      indentof(line) < current_indent()
    end

    def indentof(line)
      detab(line.slice(/\A\s*/)).length
    end

    INDENT_RE = {
      2 => /\A {2}/,
      4 => /\A {4}/,
      8 => /\A {8}/
    }

    def unindent(line, n)
      re = (INDENT_RE[n] ||= /\A {#{n}}/)
      line.sub(re, '')
    end

    #
    # I/O
    #

    def print(str)
      @result << str
    end

    def puts(str)
      @result << str
      @result << "\n" unless /\n\z/ =~ str
    end

    class LineInput
      def initialize(f)
        @f = f
        @buf = []
      end

      def inspect
        "\#<#{self.class} file=#{@f.inspect} line=#{lineno()}>"
      end

      def lineno
        @f.lineno
      end

      def gets
        return nil unless @buf
        return @buf.pop unless @buf.empty?
        line = @f.gets
        unless line
          @buf = nil
          return nil
        end
        line.rstrip
      end

      def peek
        line = gets()
        ungets line if line
        line
      end

      def ungets(line)
        @buf.push line
      end

      def next?
        peek() ? true : false
      end

      def skip_blank_lines
        n = 0
        while line = gets()
          unless line.strip.empty?
            ungets line
            return n
          end
          n += 1
        end
        n
      end

      def while_match(re)
        while line = gets()
          unless re =~ line
            ungets line
            return
          end
          yield line
        end
        nil
      end

      def until_match(re)
        while line = gets()
          if re =~ line
            ungets line
            return
          end
          yield line
        end
        nil
      end

      def until_terminator(re)
        while line = gets()
          return if re =~ line   # discard terminal line
          yield line
        end
        nil
      end
    end # LineInput

    NeedESC = /[&\"<>]/
    ESC = {
      '&' => '&amp;',
      '"' => '&quot;',
      '<' => '&lt;',
      '>' => '&gt;'
    }

    def escape_html(str)
      table = ESC   # optimize
      str.gsub(NeedESC) {|s| table[s] }
    end

    #
    # Inline
    #
    ERB_PATTERN = /<%=.*?%>/
    
    def text(str)
      s = str.split(/(#{ERB_PATTERN})/o)
      0.step(s.length, 2) do |i|
        part = s[i]
        if part
          s[i] = part.gsub(/(#{NeedESC})|(#{URI.regexp(%w(http https fttp))})/on) do
            if    ch  = $1 then ESC[ch]
            elsif url = $2 then %Q[<a href="#{escape_html(url)}">#{escape_html(url)}</a>]
            else
              raise 'must not happen'
            end
          end
        end
      end
      s.join
    end

    #
    # TextUtils
    #

    def detab(str, ts = 8)
      add = 0
      str.gsub(/\t/) {
        len = ts - ($~.begin(0) + add) % ts
        add += len - 1
        ' ' * len
      }
    end

  end # KKStyle

  class KkDiary
    include DiaryBase
    include CategorizableDiary

    def initialize( date, title, body, modified = Time::now )
      init_diary
      replace(date, title, body)
      @last_modified = modified
    end

    def style
      'kk'
    end

    def replace(date, title, body)
      set_date(date)
      set_title(title)
      @sections = []
      append(body)
    end

    def append(body, author = nil)
      body.gsub(/\r/, '').split(/^(?==[^=])/).find_all {|n| /^\A=/ =~ n}.each do |fragment|
        section = KkSection::new(fragment, author)
        @sections << section if section
      end
      @last_modified = Time::now
      self
    end

    def each_section
      @sections.each do |section|
        yield section
      end
    end

    def to_src
      src = ''
      each_section do |section|
        src << section.to_src
      end
      src
    end

    def to_html(opt, mode = :HTML)
      case mode
      when :CHTML
        to_chtml(opt)
      else
        to_html4(opt)
      end
    end

    def to_html4(opt)
      idx = 1
      r = ''
      each_section do |section|
        r << %Q[<div class="section">\n]
        if section.subtitle then
          r << %Q[<h3><a ]
          if opt['anchor'] then
            r << %Q[name="p#{'%02d' % idx}" ]
          end
          r << %Q[href="#{opt['index']}<%=anchor "#{date.strftime( '%Y%m%d' )}#p#{'%02d' % idx}" %>">#{opt['section_anchor']}</a> ]
          if opt['multi_user'] and section.author then
            r << %Q|[#{section.author}]|
          end
          r << %Q[#{section.categorized_subtitle}</h3>]
        end

        r << KKStyle.new(section.body).compile

        r << %Q[</div>]
        idx += 1
      end
      r
    end

    # TODO KkDiary#to_chtml
    def to_chtml(opt)
      to_html(opt)
    end

    def to_s
      "date=#{date.strftime('%Y%m%d')}, title=#{title}, body=[#{@sections.join('][')}]"
    end

  end # KkDiary


  class KkSection
    attr_reader :subtitle, :body, :author
    attr_reader :categories, :stripped_subtitle

    alias :body_to_html :body
    alias :subtitle_to_html :subtitle
    alias :stripped_subtitle_to_html :stripped_subtitle

    def initialize( fragment, author = nil )
      @author = author
      lines = fragment.gsub(/\A\n+/, '').split(/\n/) #remove null line at the beginning of the sentence
      subtitle, *@body = lines
      raise "fragments must have title" unless subtitle
      @subtitle = /\A=\s*/.match(subtitle).post_match
      @body = @body.join("\n") + "\n"
      @categories = get_categories
      @stripped_subtitle = strip_subtitle
    end

    def to_src
      "= #{@subtitle}\n#{@body}\n"
    end

    def to_s
      "subtitle=#{@subtitle}, body=#{@body}"
    end

    def categorized_subtitle
      @categories.collect do |c|
        %Q|<%= category_anchor("#{c}") %>|
      end.join + @stripped_subtitle.to_s
    end

    #######
    private
    #######

    def get_categories
      return [] unless @subtitle
      cat = /^(\[(.*?)\])+/.match(@subtitle).to_a[0]
      return [] unless cat
      cat.scan(/\[(.*?)\]/).collect do |c|
        c[0].split(/,/)
      end.flatten
    end

    def strip_subtitle
      return nil unless @subtitle
      r = @subtitle.sub(/\A(\[(.*?)\])+/,'')
      if r == ""
        nil
      else
        r
      end
    end
  end # KkSection

end
